@@ -11,9 +11,12 @@
 import io
 import logging
 
-import google.cloud.exceptions
-import google.cloud.storage
-import google.auth.transport.requests
+try:
+    import google.cloud.exceptions
+    import google.cloud.storage
+    import google.auth.transport.requests
+except ImportError:
+    MISSING_DEPS = True
 
 import smart_open.bytebuffer
 import smart_open.utils
@@ -109,6 +112,7 @@ def open(
         buffer_size=DEFAULT_BUFFER_SIZE,
         min_part_size=_MIN_MIN_PART_SIZE,
         client=None,  # type: google.cloud.storage.Client
+        blob_properties=None
         ):
     """Open an GCS blob for reading or writing.
 
@@ -126,10 +130,12 @@ def open(
         The minimum part size for multipart uploads.  For writing only.
     client: google.cloud.storage.Client, optional
         The GCS client to use when working with google-cloud-storage.
+    blob_properties: dict, optional
+        Set properties on blob before writing.  For writing only.
 
     """
     if mode == constants.READ_BINARY:
-        return Reader(
+        fileobj = Reader(
             bucket_id,
             blob_id,
             buffer_size=buffer_size,
@@ -137,15 +143,19 @@ def open(
             client=client,
         )
     elif mode == constants.WRITE_BINARY:
-        return Writer(
+        fileobj = Writer(
             bucket_id,
             blob_id,
             min_part_size=min_part_size,
             client=client,
+            blob_properties=blob_properties,
         )
     else:
         raise NotImplementedError('GCS support for mode %r not implemented' % mode)
 
+    fileobj.name = blob_id
+    return fileobj
+
 
 class _RawReader(object):
     """Read an GCS object."""
@@ -182,10 +192,10 @@ class _RawReader(object):
             #
             binary = b''
         elif size == -1:
-            binary = self._blob.download_as_string(start=start)
+            binary = self._blob.download_as_bytes(start=start)
         else:
             end = position + size
-            binary = self._blob.download_as_string(start=start, end=end)
+            binary = self._blob.download_as_bytes(start=start, end=end)
         return binary
 
 
@@ -208,7 +218,7 @@ class Reader(io.BufferedIOBase):
         if client is None:
             client = google.cloud.storage.Client()
 
-        self._blob = client.get_bucket(bucket).get_blob(key)  # type: google.cloud.storage.Blob
+        self._blob = client.bucket(bucket).get_blob(key)  # type: google.cloud.storage.Blob
 
         if self._blob is None:
             raise google.cloud.exceptions.NotFound('blob %s not found in %s' % (key, bucket))
@@ -390,6 +400,7 @@ class Writer(io.BufferedIOBase):
             blob,
             min_part_size=_DEFAULT_MIN_PART_SIZE,
             client=None,  # type: google.cloud.storage.Client
+            blob_properties=None,
     ):
         if client is None:
             client = google.cloud.storage.Client()
@@ -406,6 +417,10 @@ class Writer(io.BufferedIOBase):
 
         self._session = google.auth.transport.requests.AuthorizedSession(client._credentials)
 
+        if blob_properties:
+            for k, v in blob_properties.items():
+                setattr(self._blob, k, v)
+
         #
         # https://cloud.google.com/storage/docs/json_api/v1/how-tos/resumable-upload#start-resumable
         #
@@ -440,6 +455,20 @@ class Writer(io.BufferedIOBase):
         """Return True if the stream supports writing."""
         return True
 
+    def seekable(self):
+        """If False, seek(), tell() and truncate() will raise IOError.
+
+        We offer only tell support, and no seek or truncate support."""
+        return True
+
+    def seek(self, offset, whence=constants.WHENCE_START):
+        """Unsupported."""
+        raise io.UnsupportedOperation
+
+    def truncate(self, size=None):
+        """Unsupported."""
+        raise io.UnsupportedOperation
+
     def tell(self):
         """Return the current stream position."""
         return self._total_size
