@@ -13,10 +13,7 @@ import os
 import time
 import uuid
 import unittest
-try:
-    from unittest import mock
-except ImportError:
-    import mock
+from unittest import mock
 import warnings
 from collections import OrderedDict
 
@@ -72,8 +69,8 @@ class FakeBucket(object):
     def get_blob(self, blob_id):
         try:
             return self.blobs[blob_id]
-        except KeyError:
-            raise google.cloud.exceptions.NotFound('Blob {} not found'.format(blob_id))
+        except KeyError as e:
+            raise google.cloud.exceptions.NotFound('Blob {} not found'.format(blob_id)) from e
 
     def list_blobs(self):
         return list(self.blobs.values())
@@ -136,14 +133,13 @@ class FakeBucketTest(unittest.TestCase):
 
 
 class FakeBlob(object):
-    def __init__(self, name, bucket, create=True):
+    def __init__(self, name, bucket):
         self.name = name
         self._bucket = bucket  # type: FakeBucket
         self._exists = False
         self.__contents = io.BytesIO()
 
-        if create:
-            self._create_if_not_exists()
+        self._create_if_not_exists()
 
     def create_resumable_upload_session(self):
         resumeable_upload_url = RESUMABLE_SESSION_URI_TEMPLATE % dict(
@@ -158,9 +154,9 @@ class FakeBlob(object):
         self._bucket.delete_blob(self)
         self._exists = False
 
-    def download_as_string(self, start=0, end=None):
-        # mimics Google's API by returning bytes, despite the method name
-        # https://google-cloud-python.readthedocs.io/en/0.32.0/storage/blobs.html#google.cloud.storage.blob.Blob.download_as_string
+    def download_as_bytes(self, start=0, end=None):
+        # mimics Google's API by returning bytes
+        # https://googleapis.dev/python/storage/latest/blobs.html#google.cloud.storage.blob.Blob.download_as_bytes
         if end is None:
             end = self.__contents.tell()
         self.__contents.seek(start)
@@ -171,7 +167,7 @@ class FakeBlob(object):
 
     def upload_from_string(self, data):
         # mimics Google's API by accepting bytes or str, despite the method name
-        # https://google-cloud-python.readthedocs.io/en/0.32.0/storage/blobs.html#google.cloud.storage.blob.Blob.upload_from_string
+        # https://googleapis.dev/python/storage/latest/blobs.html#google.cloud.storage.blob.Blob.upload_from_string
         if isinstance(data, str):
             data = bytes(data, 'utf8')
         self.__contents = io.BytesIO(data)
@@ -215,10 +211,10 @@ class FakeBlobTest(unittest.TestCase):
         blob = FakeBlob('fake-blob', self.bucket)
         contents = b'test'
         blob.upload_from_string(contents)
-        self.assertEqual(blob.download_as_string(), b'test')
-        self.assertEqual(blob.download_as_string(start=2), b'st')
-        self.assertEqual(blob.download_as_string(end=2), b'te')
-        self.assertEqual(blob.download_as_string(start=2, end=3), b's')
+        self.assertEqual(blob.download_as_bytes(), b'test')
+        self.assertEqual(blob.download_as_bytes(start=2), b'st')
+        self.assertEqual(blob.download_as_bytes(end=2), b'te')
+        self.assertEqual(blob.download_as_bytes(start=2, end=3), b's')
 
     def test_size(self):
         blob = FakeBlob('fake-blob', self.bucket)
@@ -246,8 +242,8 @@ class FakeClient(object):
     def bucket(self, bucket_id):
         try:
             return self.__buckets[bucket_id]
-        except KeyError:
-            raise google.cloud.exceptions.NotFound('Bucket %s not found' % bucket_id)
+        except KeyError as e:
+            raise google.cloud.exceptions.NotFound('Bucket %s not found' % bucket_id) from e
 
     def create_bucket(self, bucket_id):
         bucket = FakeBucket(self, bucket_id)
@@ -373,7 +369,7 @@ class FakeAuthorizedSessionTest(unittest.TestCase):
         response = self.session.put(self.upload_url, data, headers=headers)
         self.assertIn(response.status_code, smart_open.gcs._UPLOAD_INCOMPLETE_STATUS_CODES)
         self.session._blob_with_url(self.upload_url, self.client)
-        blob_contents = self.blob.download_as_string()
+        blob_contents = self.blob.download_as_bytes()
         self.assertEqual(blob_contents, b'')
 
     def test_finished_put_writes_to_blob(self):
@@ -385,7 +381,7 @@ class FakeAuthorizedSessionTest(unittest.TestCase):
         response = self.session.put(self.upload_url, data, headers=headers)
         self.assertEqual(response.status_code, 200)
         self.session._blob_with_url(self.upload_url, self.client)
-        blob_contents = self.blob.download_as_string()
+        blob_contents = self.blob.download_as_bytes()
         data.seek(0)
         self.assertEqual(blob_contents, data.read())
 
@@ -805,6 +801,18 @@ class WriterTest(unittest.TestCase):
 
         self.assertEqual(output, [])
 
+    def test_write_05(self):
+        """Do blob_properties get applied?"""
+        smart_open_write = smart_open.gcs.Writer(BUCKET_NAME, WRITE_BLOB_NAME,
+            blob_properties={
+                "content_type": "random/x-test",
+                "content_encoding": "coded"
+            }
+        )
+        with smart_open_write as fout:  # noqa
+            assert fout._blob.content_type == "random/x-test"
+            assert fout._blob.content_encoding == "coded"
+
     def test_gzip(self):
         expected = u'а не спеть ли мне песню... о любви'.encode('utf-8')
         with smart_open.gcs.Writer(BUCKET_NAME, WRITE_BLOB_NAME) as fout:
@@ -889,13 +897,26 @@ class OpenTest(unittest.TestCase):
         """read should never return None."""
         test_string = u"ветер по морю гуляет..."
         with smart_open.gcs.open(BUCKET_NAME, BLOB_NAME, "wb") as fout:
+            self.assertEqual(fout.name, BLOB_NAME)
             fout.write(test_string.encode('utf8'))
 
         r = smart_open.gcs.open(BUCKET_NAME, BLOB_NAME, "rb")
+        self.assertEqual(r.name, BLOB_NAME)
         self.assertEqual(r.read(), test_string.encode("utf-8"))
         self.assertEqual(r.read(), b"")
         self.assertEqual(r.read(), b"")
 
+    def test_round_trip(self):
+        test_string = u"ветер по морю гуляет..."
+        url = 'gs://%s/%s' % (BUCKET_NAME, BLOB_NAME)
+        with smart_open.open(url, "w", encoding='utf-8') as fout:
+            fout.write(test_string)
+
+        with smart_open.open(url, encoding='utf-8') as fin:
+            actual = fin.read()
+
+        self.assertEqual(test_string, actual)
+
 
 class MakeRangeStringTest(unittest.TestCase):
     def test_no_stop(self):
