@@ -5,19 +5,24 @@
 # This code is distributed under the terms and conditions
 # from the MIT License (MIT).
 #
+from collections import defaultdict
 import gzip
 import io
 import logging
 import os
+import tempfile
 import time
 import unittest
 import warnings
+from contextlib import contextmanager
+from unittest import mock
+import sys
 
-import boto.s3.bucket
 import boto3
 import botocore.client
-import mock
+import botocore.endpoint
 import moto
+import pytest
 
 import smart_open
 import smart_open.s3
@@ -33,6 +38,13 @@ KEY_NAME = 'test-key'
 WRITE_KEY_NAME = 'test-write-key'
 ENABLE_MOTO_SERVER = os.environ.get("SO_ENABLE_MOTO_SERVER") == "1"
 
+#
+# This is a hack to keep moto happy
+# See https://github.com/spulec/moto/issues/1941
+#
+os.environ["AWS_ACCESS_KEY_ID"] = "test"
+os.environ["AWS_SECRET_ACCESS_KEY"] = "test"
+
 
 logger = logging.getLogger(__name__)
 
@@ -79,33 +91,137 @@ def ignore_resource_warnings():
     warnings.filterwarnings("ignore", category=ResourceWarning, message="unclosed.*<ssl.SSLSocket.*>")  # noqa
 
 
+@contextmanager
+def patch_invalid_range_response(actual_size):
+    """ Work around a bug in moto (https://github.com/spulec/moto/issues/2981) where the
+     API response doesn't match when requesting an invalid range of bytes from an S3 GetObject. """
+    _real_get = smart_open.s3._get
+
+    def mock_get(*args, **kwargs):
+        try:
+            return _real_get(*args, **kwargs)
+        except IOError as ioe:
+            error_response = smart_open.s3._unwrap_ioerror(ioe)
+            if error_response and error_response.get('Message') == 'Requested Range Not Satisfiable':
+                error_response['ActualObjectSize'] = actual_size
+                error_response['Code'] = 'InvalidRange'
+                error_response['Message'] = 'The requested range is not satisfiable'
+            raise
+
+    with mock.patch('smart_open.s3._get', new=mock_get):
+        yield
+
+
+class BaseTest(unittest.TestCase):
+    @contextmanager
+    def assertApiCalls(self, **expected_api_calls):
+        """ Track calls to S3 in self.api_calls by patching botocore.endpoint.Endpoint.make_request. """
+        _real_make_request = botocore.endpoint.Endpoint.make_request
+        api_calls = defaultdict(int)
+
+        def mock_make_request(self, operation_model, *args, **kwargs):
+            api_calls[operation_model.name] += 1
+            return _real_make_request(self, operation_model, *args, **kwargs)
+
+        patcher = mock.patch('botocore.endpoint.Endpoint.make_request', new=mock_make_request)
+        patcher.start()
+        try:
+            yield api_calls
+            self.assertDictEqual(expected_api_calls, api_calls)
+        finally:
+            patcher.stop()
+
+
 @unittest.skipUnless(
     ENABLE_MOTO_SERVER,
     'The test case needs a Moto server running on the local 5000 port.'
 )
 class SeekableRawReaderTest(unittest.TestCase):
-
     def setUp(self):
         self._body = b'123456'
         self._local_resource = boto3.resource('s3', endpoint_url='http://localhost:5000')
         self._local_resource.Bucket(BUCKET_NAME).create()
         self._local_resource.Object(BUCKET_NAME, KEY_NAME).put(Body=self._body)
+        self._local_client = boto3.client('s3', endpoint_url='http://localhost:5000')
 
     def tearDown(self):
         self._local_resource.Object(BUCKET_NAME, KEY_NAME).delete()
         self._local_resource.Bucket(BUCKET_NAME).delete()
 
     def test_read_from_a_closed_body(self):
-        obj = self._local_resource.Object(BUCKET_NAME, KEY_NAME)
-        content_length = obj.content_length
-        reader = smart_open.s3._SeekableRawReader(obj, content_length)
+        reader = smart_open.s3._SeekableRawReader(self._local_client, BUCKET_NAME, KEY_NAME)
         self.assertEqual(reader.read(1), b'1')
         reader._body.close()
         self.assertEqual(reader.read(2), b'23')
 
 
+class CrapStream(io.BytesIO):
+    """Raises an exception on every second read call."""
+    def __init__(self, *args, modulus=2, **kwargs):
+        super().__init__(*args, **kwargs)
+        self._count = 0
+        self._modulus = modulus
+
+    def read(self, size=-1):
+        self._count += 1
+        if self._count % self._modulus == 0:
+            raise botocore.exceptions.BotoCoreError()
+        the_bytes = super().read(size)
+        return the_bytes
+
+
+class CrapClient:
+    def __init__(self, data, modulus=2):
+        self._datasize = len(data)
+        self._body = CrapStream(data, modulus=modulus)
+
+    def get_object(self, *args, **kwargs):
+        return {
+            'ActualObjectSize': self._datasize,
+            'ContentLength': self._datasize,
+            'ContentRange': 'bytes 0-%d/%d' % (self._datasize, self._datasize),
+            'Body': self._body,
+            'ResponseMetadata': {'RetryAttempts': 1},
+        }
+
+
+class IncrementalBackoffTest(unittest.TestCase):
+    def test_every_read_fails(self):
+        reader = smart_open.s3._SeekableRawReader(CrapClient(b'hello', 1), 'bucket', 'key')
+        with mock.patch('time.sleep') as mock_sleep:
+            with self.assertRaises(IOError):
+                reader.read()
+
+            #
+            # Make sure our incremental backoff is actually happening here.
+            #
+            mock_sleep.assert_has_calls([mock.call(s) for s in (1, 2, 4, 8, 16)])
+
+    def test_every_second_read_fails(self):
+        """Can we read from a stream that raises exceptions from time to time?"""
+        reader = smart_open.s3._SeekableRawReader(CrapClient(b'hello'), 'bucket', 'key')
+        with mock.patch('time.sleep') as mock_sleep:
+            assert reader.read(1) == b'h'
+            mock_sleep.assert_not_called()
+
+            assert reader.read(1) == b'e'
+            mock_sleep.assert_called_with(1)
+            mock_sleep.reset_mock()
+
+            assert reader.read(1) == b'l'
+            mock_sleep.reset_mock()
+
+            assert reader.read(1) == b'l'
+            mock_sleep.assert_called_with(1)
+            mock_sleep.reset_mock()
+
+            assert reader.read(1) == b'o'
+            mock_sleep.assert_called_with(1)
+            mock_sleep.reset_mock()
+
+
 @moto.mock_s3
-class SeekableBufferedInputBaseTest(unittest.TestCase):
+class ReaderTest(BaseTest):
     def setUp(self):
         # lower the multipart upload size, to speed up these tests
         self.old_min_part_size = smart_open.s3.DEFAULT_MIN_PART_SIZE
@@ -113,6 +229,8 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
 
         ignore_resource_warnings()
 
+        super().setUp()
+
     def tearDown(self):
         smart_open.s3.DEFAULT_MIN_PART_SIZE = self.old_min_part_size
         cleanup_bucket()
@@ -124,17 +242,19 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
         put_to_bucket(contents=expected)
 
         # connect to fake s3 and read from the fake key we filled above
-        fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
-        output = [line.rstrip(b'\n') for line in fin]
+        with self.assertApiCalls(GetObject=1):
+            fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME)
+            output = [line.rstrip(b'\n') for line in fin]
         self.assertEqual(output, expected.split(b'\n'))
 
     def test_iter_context_manager(self):
         # same thing but using a context manager
         expected = u"hello wořld\nhow are you?".encode('utf8')
         put_to_bucket(contents=expected)
-        with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME) as fin:
-            output = [line.rstrip(b'\n') for line in fin]
-            self.assertEqual(output, expected.split(b'\n'))
+        with self.assertApiCalls(GetObject=1):
+            with smart_open.s3.Reader(BUCKET_NAME, KEY_NAME) as fin:
+                output = [line.rstrip(b'\n') for line in fin]
+        self.assertEqual(output, expected.split(b'\n'))
 
     def test_read(self):
         """Are S3 files read correctly?"""
@@ -142,68 +262,89 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
         put_to_bucket(contents=content)
         logger.debug('content: %r len: %r', content, len(content))
 
-        fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
-        self.assertEqual(content[:6], fin.read(6))
-        self.assertEqual(content[6:14], fin.read(8))  # ř is 2 bytes
-        self.assertEqual(content[14:], fin.read())  # read the rest
+        with self.assertApiCalls(GetObject=1):
+            fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME)
+            self.assertEqual(content[:6], fin.read(6))
+            self.assertEqual(content[6:14], fin.read(8))  # ř is 2 bytes
+            self.assertEqual(content[14:], fin.read())  # read the rest
 
     def test_seek_beginning(self):
         """Does seeking to the beginning of S3 files work correctly?"""
         content = u"hello wořld\nhow are you?".encode('utf8')
         put_to_bucket(contents=content)
 
-        fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
-        self.assertEqual(content[:6], fin.read(6))
-        self.assertEqual(content[6:14], fin.read(8))  # ř is 2 bytes
+        with self.assertApiCalls(GetObject=1):
+            fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME)
+            self.assertEqual(content[:6], fin.read(6))
+            self.assertEqual(content[6:14], fin.read(8))  # ř is 2 bytes
 
-        fin.seek(0)
-        self.assertEqual(content, fin.read())  # no size given => read whole file
+        with self.assertApiCalls(GetObject=1):
+            fin.seek(0)
+            self.assertEqual(content, fin.read())  # no size given => read whole file
 
-        fin.seek(0)
-        self.assertEqual(content, fin.read(-1))  # same thing
+        with self.assertApiCalls(GetObject=1):
+            fin.seek(0)
+            self.assertEqual(content, fin.read(-1))  # same thing
 
     def test_seek_start(self):
         """Does seeking from the start of S3 files work correctly?"""
         content = u"hello wořld\nhow are you?".encode('utf8')
         put_to_bucket(contents=content)
 
-        fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
-        seek = fin.seek(6)
-        self.assertEqual(seek, 6)
-        self.assertEqual(fin.tell(), 6)
-        self.assertEqual(fin.read(6), u'wořld'.encode('utf-8'))
+        with self.assertApiCalls(GetObject=1):
+            fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME, defer_seek=True)
+            seek = fin.seek(6)
+            self.assertEqual(seek, 6)
+            self.assertEqual(fin.tell(), 6)
+            self.assertEqual(fin.read(6), u'wořld'.encode('utf-8'))
 
     def test_seek_current(self):
         """Does seeking from the middle of S3 files work correctly?"""
         content = u"hello wořld\nhow are you?".encode('utf8')
         put_to_bucket(contents=content)
 
-        fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
-        self.assertEqual(fin.read(5), b'hello')
-        seek = fin.seek(1, whence=smart_open.constants.WHENCE_CURRENT)
-        self.assertEqual(seek, 6)
-        self.assertEqual(fin.read(6), u'wořld'.encode('utf-8'))
+        with self.assertApiCalls(GetObject=1):
+            fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME)
+            self.assertEqual(fin.read(5), b'hello')
+
+        with self.assertApiCalls(GetObject=1):
+            seek = fin.seek(1, whence=smart_open.constants.WHENCE_CURRENT)
+            self.assertEqual(seek, 6)
+            self.assertEqual(fin.read(6), u'wořld'.encode('utf-8'))
 
     def test_seek_end(self):
         """Does seeking from the end of S3 files work correctly?"""
         content = u"hello wořld\nhow are you?".encode('utf8')
         put_to_bucket(contents=content)
 
-        fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
-        seek = fin.seek(-4, whence=smart_open.constants.WHENCE_END)
-        self.assertEqual(seek, len(content) - 4)
-        self.assertEqual(fin.read(), b'you?')
+        with self.assertApiCalls(GetObject=1):
+            fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME, defer_seek=True)
+            seek = fin.seek(-4, whence=smart_open.constants.WHENCE_END)
+            self.assertEqual(seek, len(content) - 4)
+            self.assertEqual(fin.read(), b'you?')
+
+    def test_seek_past_end(self):
+        content = u"hello wořld\nhow are you?".encode('utf8')
+        put_to_bucket(contents=content)
+
+        with self.assertApiCalls(GetObject=1), patch_invalid_range_response(str(len(content))):
+            fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME, defer_seek=True)
+            seek = fin.seek(60)
+            self.assertEqual(seek, len(content))
 
     def test_detect_eof(self):
         content = u"hello wořld\nhow are you?".encode('utf8')
         put_to_bucket(contents=content)
 
-        fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
-        fin.read()
-        eof = fin.tell()
-        self.assertEqual(eof, len(content))
-        fin.seek(0, whence=smart_open.constants.WHENCE_END)
-        self.assertEqual(eof, fin.tell())
+        with self.assertApiCalls(GetObject=1):
+            fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME)
+            fin.read()
+            eof = fin.tell()
+            self.assertEqual(eof, len(content))
+            fin.seek(0, whence=smart_open.constants.WHENCE_END)
+            self.assertEqual(eof, fin.tell())
+            fin.seek(eof)
+            self.assertEqual(eof, fin.tell())
 
     def test_read_gzip(self):
         expected = u'раcцветали яблони и груши, поплыли туманы над рекой...'.encode('utf-8')
@@ -216,7 +357,7 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
         #
         # Make sure we're reading things correctly.
         #
-        with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME) as fin:
+        with smart_open.s3.Reader(BUCKET_NAME, KEY_NAME) as fin:
             self.assertEqual(fin.read(), buf.getvalue())
 
         #
@@ -227,9 +368,10 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
             self.assertEqual(zipfile.read(), expected)
 
         logger.debug('starting actual test')
-        with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME) as fin:
-            with gzip.GzipFile(fileobj=fin) as zipfile:
-                actual = zipfile.read()
+        with self.assertApiCalls(GetObject=1):
+            with smart_open.s3.Reader(BUCKET_NAME, KEY_NAME) as fin:
+                with gzip.GzipFile(fileobj=fin) as zipfile:
+                    actual = zipfile.read()
 
         self.assertEqual(expected, actual)
 
@@ -237,13 +379,14 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
         content = b'englishman\nin\nnew\nyork\n'
         put_to_bucket(contents=content)
 
-        with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME) as fin:
-            fin.readline()
-            self.assertEqual(fin.tell(), content.index(b'\n')+1)
+        with self.assertApiCalls(GetObject=2):
+            with smart_open.s3.Reader(BUCKET_NAME, KEY_NAME) as fin:
+                fin.readline()
+                self.assertEqual(fin.tell(), content.index(b'\n')+1)
 
-            fin.seek(0)
-            actual = list(fin)
-            self.assertEqual(fin.tell(), len(content))
+                fin.seek(0)
+                actual = list(fin)
+                self.assertEqual(fin.tell(), len(content))
 
         expected = [b'englishman\n', b'in\n', b'new\n', b'york\n']
         self.assertEqual(expected, actual)
@@ -252,8 +395,9 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
         content = b'englishman\nin\nnew\nyork\n'
         put_to_bucket(contents=content)
 
-        with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME, buffer_size=8) as fin:
-            actual = list(fin)
+        with self.assertApiCalls(GetObject=1):
+            with smart_open.s3.Reader(BUCKET_NAME, KEY_NAME, buffer_size=8) as fin:
+                actual = list(fin)
 
         expected = [b'englishman\n', b'in\n', b'new\n', b'york\n']
         self.assertEqual(expected, actual)
@@ -262,8 +406,10 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
         content = b'englishman\nin\nnew\nyork\n'
         put_to_bucket(contents=content)
 
-        with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME) as fin:
-            data = fin.read(0)
+        with self.assertApiCalls():
+            # set defer_seek to verify that read(0) doesn't trigger an unnecessary API call
+            with smart_open.s3.Reader(BUCKET_NAME, KEY_NAME, defer_seek=True) as fin:
+                data = fin.read(0)
 
         self.assertEqual(data, b'')
 
@@ -271,8 +417,10 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
         contents = b'the spice melange\n'
         put_to_bucket(contents=contents)
 
-        with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME) as fin:
-            returned_obj = fin.to_boto3()
+        with self.assertApiCalls():
+            # set defer_seek to verify that to_boto3() doesn't trigger an unnecessary API call
+            with smart_open.s3.Reader(BUCKET_NAME, KEY_NAME, defer_seek=True) as fin:
+                returned_obj = fin.to_boto3(boto3.resource('s3'))
 
         boto3_body = returned_obj.get()['Body'].read()
         self.assertEqual(contents, boto3_body)
@@ -280,10 +428,35 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
     def test_binary_iterator(self):
         expected = u"выйду ночью в поле с конём".encode('utf-8').split(b' ')
         put_to_bucket(contents=b"\n".join(expected))
-        with smart_open.s3.open(BUCKET_NAME, KEY_NAME, 'rb') as fin:
-            actual = [line.rstrip() for line in fin]
+        with self.assertApiCalls(GetObject=1):
+            with smart_open.s3.open(BUCKET_NAME, KEY_NAME, 'rb') as fin:
+                actual = [line.rstrip() for line in fin]
         self.assertEqual(expected, actual)
 
+    def test_defer_seek(self):
+        content = b'englishman\nin\nnew\nyork\n'
+        put_to_bucket(contents=content)
+
+        with self.assertApiCalls():
+            fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME, defer_seek=True)
+        with self.assertApiCalls(GetObject=1):
+            self.assertEqual(fin.read(), content)
+
+        with self.assertApiCalls():
+            fin = smart_open.s3.Reader(BUCKET_NAME, KEY_NAME, defer_seek=True)
+        with self.assertApiCalls(GetObject=1):
+            fin.seek(10)
+            self.assertEqual(fin.read(), content[10:])
+
+    def test_read_empty_file(self):
+        put_to_bucket(contents=b'')
+
+        with self.assertApiCalls(GetObject=1), patch_invalid_range_response('0'):
+            with smart_open.s3.Reader(BUCKET_NAME, KEY_NAME) as fin:
+                data = fin.read()
+
+        self.assertEqual(data, b'')
+
 
 @moto.mock_s3
 class MultipartWriterTest(unittest.TestCase):
@@ -306,7 +479,7 @@ class MultipartWriterTest(unittest.TestCase):
             fout.write(test_string)
 
         # read key and test content
-        output = list(smart_open.smart_open("s3://{}/{}".format(BUCKET_NAME, WRITE_KEY_NAME), "rb"))
+        output = list(smart_open.s3.open(BUCKET_NAME, WRITE_KEY_NAME, "rb"))
 
         self.assertEqual(output, [test_string])
 
@@ -348,7 +521,7 @@ class MultipartWriterTest(unittest.TestCase):
             self.assertEqual(fout._total_parts, 1)
 
         # read back the same key and check its content
-        output = list(smart_open.smart_open("s3://{}/{}".format(BUCKET_NAME, WRITE_KEY_NAME)))
+        output = list(smart_open.s3.open(BUCKET_NAME, WRITE_KEY_NAME, 'rb'))
         self.assertEqual(output, [b"testtest\n", b"test"])
 
     def test_write_04(self):
@@ -358,7 +531,8 @@ class MultipartWriterTest(unittest.TestCase):
             pass
 
         # read back the same key and check its content
-        output = list(smart_open.smart_open("s3://{}/{}".format(BUCKET_NAME, WRITE_KEY_NAME)))
+        with patch_invalid_range_response('0'):
+            output = list(smart_open.s3.open(BUCKET_NAME, WRITE_KEY_NAME, 'rb'))
 
         self.assertEqual(output, [])
 
@@ -368,7 +542,7 @@ class MultipartWriterTest(unittest.TestCase):
             with gzip.GzipFile(fileobj=fout, mode='w') as zipfile:
                 zipfile.write(expected)
 
-        with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, WRITE_KEY_NAME) as fin:
+        with smart_open.s3.Reader(BUCKET_NAME, WRITE_KEY_NAME) as fin:
             with gzip.GzipFile(fileobj=fin) as zipfile:
                 actual = zipfile.read()
 
@@ -385,7 +559,7 @@ class MultipartWriterTest(unittest.TestCase):
             with io.BufferedWriter(fout) as sub_out:
                 sub_out.write(expected.encode('utf-8'))
 
-        with smart_open.smart_open("s3://{}/{}".format(BUCKET_NAME, WRITE_KEY_NAME)) as fin:
+        with smart_open.s3.open(BUCKET_NAME, WRITE_KEY_NAME, 'rb') as fin:
             with io.TextIOWrapper(fin, encoding='utf-8') as text:
                 actual = text.read()
 
@@ -421,11 +595,24 @@ class MultipartWriterTest(unittest.TestCase):
 
         with smart_open.s3.open(BUCKET_NAME, KEY_NAME, 'wb') as fout:
             fout.write(contents)
-            returned_obj = fout.to_boto3()
+            returned_obj = fout.to_boto3(boto3.resource('s3'))
 
         boto3_body = returned_obj.get()['Body'].read()
         self.assertEqual(contents, boto3_body)
 
+    def test_writebuffer(self):
+        """Does the MultipartWriter support writing to a custom buffer?"""
+        contents = b'get ready for a surprise'
+
+        with tempfile.NamedTemporaryFile(mode='rb+') as f:
+            with smart_open.s3.MultipartWriter(BUCKET_NAME, WRITE_KEY_NAME, writebuffer=f) as fout:
+                fout.write(contents)
+
+            with smart_open.s3.open(BUCKET_NAME, WRITE_KEY_NAME, 'rb') as fin:
+                actual = fin.read()
+
+            assert actual == contents
+
 
 @moto.mock_s3
 class SinglepartWriterTest(unittest.TestCase):
@@ -448,7 +635,7 @@ class SinglepartWriterTest(unittest.TestCase):
             fout.write(test_string)
 
         # read key and test content
-        output = list(smart_open.smart_open("s3://{}/{}".format(BUCKET_NAME, WRITE_KEY_NAME), "rb"))
+        output = list(smart_open.s3.open(BUCKET_NAME, WRITE_KEY_NAME, "rb"))
 
         self.assertEqual(output, [test_string])
 
@@ -480,8 +667,8 @@ class SinglepartWriterTest(unittest.TestCase):
             pass
 
         # read back the same key and check its content
-        output = list(smart_open.smart_open("s3://{}/{}".format(BUCKET_NAME, WRITE_KEY_NAME)))
-
+        with patch_invalid_range_response('0'):
+            output = list(smart_open.s3.open(BUCKET_NAME, WRITE_KEY_NAME, 'rb'))
         self.assertEqual(output, [])
 
     def test_buffered_writer_wrapper_works(self):
@@ -495,7 +682,7 @@ class SinglepartWriterTest(unittest.TestCase):
             with io.BufferedWriter(fout) as sub_out:
                 sub_out.write(expected.encode('utf-8'))
 
-        with smart_open.smart_open("s3://{}/{}".format(BUCKET_NAME, WRITE_KEY_NAME)) as fin:
+        with smart_open.s3.open(BUCKET_NAME, WRITE_KEY_NAME, 'rb') as fin:
             with io.TextIOWrapper(fin, encoding='utf-8') as text:
                 actual = text.read()
 
@@ -521,6 +708,19 @@ class SinglepartWriterTest(unittest.TestCase):
         fout.flush()
         fout.close()
 
+    def test_writebuffer(self):
+        """Does the SinglepartWriter support writing to a custom buffer?"""
+        contents = b'get ready for a surprise'
+
+        with tempfile.NamedTemporaryFile(mode='rb+') as f:
+            with smart_open.s3.SinglepartWriter(BUCKET_NAME, WRITE_KEY_NAME, writebuffer=f) as fout:
+                fout.write(contents)
+
+            with smart_open.s3.open(BUCKET_NAME, WRITE_KEY_NAME, 'rb') as fin:
+                actual = fin.read()
+
+            assert actual == contents
+
 
 ARBITRARY_CLIENT_ERROR = botocore.client.ClientError(error_response={}, operation_name='bar')
 
@@ -533,23 +733,39 @@ class IterBucketTest(unittest.TestCase):
     def tearDown(self):
         cleanup_bucket()
 
+    @pytest.mark.skipif(condition=sys.platform == 'win32', reason="does not run on windows")
+    @pytest.mark.xfail(
+        condition=sys.platform == 'darwin',
+        reason="MacOS uses spawn rather than fork for multiprocessing",
+    )
     def test_iter_bucket(self):
         populate_bucket()
         results = list(smart_open.s3.iter_bucket(BUCKET_NAME))
         self.assertEqual(len(results), 10)
 
+    def test_deprecated_top_level_s3_iter_bucket(self):
+        populate_bucket()
+        with self.assertLogs(smart_open.logger.name, level='WARN') as cm:
+            # invoking once will generate a warning
+            smart_open.s3_iter_bucket(BUCKET_NAME)
+            # invoking again will not (to reduce spam)
+            smart_open.s3_iter_bucket(BUCKET_NAME)
+            # verify only one output
+            assert len(cm.output) == 1
+            # verify the suggested new import is in the warning
+            assert "from smart_open.s3 import iter_bucket as s3_iter_bucket" in cm.output[0]
+
+    @pytest.mark.skipif(condition=sys.platform == 'win32', reason="does not run on windows")
+    @pytest.mark.xfail(
+        condition=sys.platform == 'darwin',
+        reason="MacOS uses spawn rather than fork for multiprocessing",
+    )
     def test_accepts_boto3_bucket(self):
         populate_bucket()
         bucket = boto3.resource('s3').Bucket(BUCKET_NAME)
         results = list(smart_open.s3.iter_bucket(bucket))
         self.assertEqual(len(results), 10)
 
-    def test_accepts_boto_bucket(self):
-        populate_bucket()
-        bucket = boto.s3.bucket.Bucket(name=BUCKET_NAME)
-        results = list(smart_open.s3.iter_bucket(bucket))
-        self.assertEqual(len(results), 10)
-
     def test_list_bucket(self):
         num_keys = 10
         populate_bucket()
@@ -568,47 +784,17 @@ class IterBucketTest(unittest.TestCase):
         expected = ['key_%d' % x for x in range(num_keys)]
         self.assertEqual(sorted(keys), sorted(expected))
 
-    def test_old(self):
-        """Does s3_iter_bucket work correctly?"""
-        #
-        # Use an old-school boto Bucket class for historical reasons.
-        #
-        mybucket = boto.s3.bucket.Bucket(name=BUCKET_NAME)
-
-        # first, create some keys in the bucket
-        expected = {}
-        for key_no in range(42):
-            key_name = "mykey%s" % key_no
-            with smart_open.smart_open("s3://%s/%s" % (BUCKET_NAME, key_name), 'wb') as fout:
-                content = '\n'.join("line%i%i" % (key_no, line_no) for line_no in range(10)).encode('utf8')
-                fout.write(content)
-                expected[key_name] = content
-
-        # read all keys + their content back, in parallel, using s3_iter_bucket
-        result = {}
-        for k, c in smart_open.s3.iter_bucket(mybucket):
-            result[k] = c
-        self.assertEqual(expected, result)
-
-        # read some of the keys back, in parallel, using s3_iter_bucket
-        result = {}
-        for k, c in smart_open.s3.iter_bucket(mybucket, accept_key=lambda fname: fname.endswith('4')):
-            result[k] = c
-        self.assertEqual(result, dict((k, c) for k, c in expected.items() if k.endswith('4')))
-
-        # read some of the keys back, in parallel, using s3_iter_bucket
-        result = dict(smart_open.s3.iter_bucket(mybucket, key_limit=10))
-        self.assertEqual(len(result), min(len(expected), 10))
-
-        for workers in [1, 4, 8, 16, 64]:
-            result = {}
-            for k, c in smart_open.s3.iter_bucket(mybucket):
-                result[k] = c
-            self.assertEqual(result, expected)
-
 
 @moto.mock_s3
-@unittest.skipIf(not smart_open.concurrency._CONCURRENT_FUTURES, 'concurrent.futures unavailable')
+@pytest.mark.skipif(
+    condition=not smart_open.concurrency._CONCURRENT_FUTURES,
+    reason='concurrent.futures unavailable',
+)
+@pytest.mark.skipif(condition=sys.platform == 'win32', reason="does not run on windows")
+@pytest.mark.xfail(
+    condition=sys.platform == 'darwin',
+    reason="MacOS uses spawn rather than fork for multiprocessing",
+)
 class IterBucketConcurrentFuturesTest(unittest.TestCase):
     def setUp(self):
         self.old_flag_multi = smart_open.concurrency._MULTIPROCESSING
@@ -630,7 +816,15 @@ class IterBucketConcurrentFuturesTest(unittest.TestCase):
 
 
 @moto.mock_s3
-@unittest.skipIf(not smart_open.concurrency._MULTIPROCESSING, 'multiprocessing unavailable')
+@pytest.mark.skipif(
+    condition=not smart_open.concurrency._MULTIPROCESSING,
+    reason='multiprocessing unavailable',
+)
+@pytest.mark.skipif(condition=sys.platform == 'win32', reason="does not run on windows")
+@pytest.mark.xfail(
+    condition=sys.platform == 'darwin',
+    reason="MacOS uses spawn rather than fork for multiprocessing",
+)
 class IterBucketMultiprocessingTest(unittest.TestCase):
     def setUp(self):
         self.old_flag_concurrent = smart_open.concurrency._CONCURRENT_FUTURES
@@ -768,9 +962,6 @@ class OpenTest(unittest.TestCase):
 
 
 def populate_bucket(num_keys=10):
-    # fake (or not) connection, bucket and key
-    logger.debug('%r', locals())
-
     s3 = boto3.resource('s3')
     for key_number in range(num_keys):
         key_name = 'key_%d' % key_number
@@ -794,6 +985,70 @@ class RetryIfFailedTest(unittest.TestCase):
         self.assertEqual(partial.call_count, 3)
 
 
+@moto.mock_s3()
+def test_client_propagation_singlepart():
+    """Does the client parameter make it from the caller to Boto3?"""
+    #
+    # Not sure why we need to create the bucket here, as setUpModule should
+    # have done that for us by now.
+    #
+    session = boto3.Session()
+    resource = session.resource('s3')
+    bucket = resource.create_bucket(Bucket=BUCKET_NAME)
+    bucket.wait_until_exists()
+
+    client = session.client('s3')
+
+    with smart_open.s3.open(
+        BUCKET_NAME,
+        WRITE_KEY_NAME,
+        mode='wb',
+        client=client,
+        multipart_upload=False,
+    ) as writer:
+        assert writer._client.client == client
+        assert id(writer._client.client) == id(client)
+
+
+@moto.mock_s3()
+def test_client_propagation_multipart():
+    """Does the resource parameter make it from the caller to Boto3?"""
+    session = boto3.Session()
+    resource = session.resource('s3')
+    bucket = resource.create_bucket(Bucket=BUCKET_NAME)
+    bucket.wait_until_exists()
+
+    client = session.client('s3')
+
+    with smart_open.s3.open(
+        BUCKET_NAME,
+        WRITE_KEY_NAME,
+        mode='wb',
+        client=client,
+        multipart_upload=True,
+    ) as writer:
+        assert writer._client.client == client
+        assert id(writer._client.client) == id(client)
+
+
+@moto.mock_s3()
+def test_resource_propagation_reader():
+    """Does the resource parameter make it from the caller to Boto3?"""
+    session = boto3.Session()
+    resource = session.resource('s3')
+    bucket = resource.create_bucket(Bucket=BUCKET_NAME)
+    bucket.wait_until_exists()
+
+    client = session.client('s3')
+
+    with smart_open.s3.open(BUCKET_NAME, WRITE_KEY_NAME, mode='wb') as writer:
+        writer.write(b'hello world')
+
+    with smart_open.s3.open(BUCKET_NAME, WRITE_KEY_NAME, mode='rb', client=client) as reader:
+        assert reader._client.client == client
+        assert id(reader._client.client) == id(client)
+
+
 if __name__ == '__main__':
     logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=logging.INFO)
     unittest.main()
